"""
    String starts with conditions
"""

from marshmallow import post_load

from .base import StringCondition, StringConditionSchema


class StartsWith(StringCondition):
    """
        Condition for string `self.what` starts with `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if self.case_insensitive:
                    if self.what.lower().startswith(i.lower()):
                        return True
                else:
                    if self.what.startswith(i):
                        return True
            return False
        else:
            for i in self.values:
                if self.case_insensitive:
                    if not self.what.lower().startswith(i.lower()):
                        return False
                else:
                    if not self.what.startswith(i):
                        return False
            return True


class StartsWithSchema(StringConditionSchema):
    """
        JSON schema for starts with string condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return StartsWith(**data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
               ..
              ( '`<
               )(
        ( ----'  '.
        (         ;
         (_______,' 
    ~^~^~^~^~^~^~^~^~^~^~
    Author: thong
    Company: M O B I O
    Date Created: 14/03/2023
"""
from datetime import datetime

from .adapter import DataAdapter


class ElasticsearchAdapter(DataAdapter):

    def execute_query(self, query):
        pass

    def build_query(self, policy):
        """
        :param policy:
        [{
            "effect": "allow",
            "condition": [{
                "operator": "StringEquals",
                "field": "user:staff_code",
                "values": ["A123456"],
                "qualifier": "ForAnyValue", // ForAllValues
                "if_exists": True,
                "ignore_case": True,
            }]
        }]
        :return:
        """
        must = []
        must_not = []
        should = []
        for pol in policy:
            conditions = pol.get("condition")
            effect = pol.get("effect")
            for condition in conditions:
                must_i, must_not_i, should_i = self.convert_operator_to_query(condition)
                if effect == "allow":
                    if must_i:
                        must.extend(must_i)
                    if must_not_i:
                        must_not.extend(must_not_i)
                    if should_i:
                        should.extend(should_i)
                else:
                    if must_i:
                        must_not.extend(must_i)
                    if must_not_i:
                        must.extend(must_not_i)
                    if should_i:
                        must_not.append({
                            "bool": {
                                "should": should_i
                            }
                        })

        return {
            "must": must,
            "must_not": must_not,
            "should": should
        }

    def convert_operator_to_query(self, filter):
        must = []
        must_not = []
        should = []

        operator = filter.get("operator")
        field = filter.get("field")
        values = filter.get("values")
        qualifier = filter.get("qualifier")  # default ForAllValues
        if not field or not operator:
            return must, must_not, should

        field_keys = field.split(":")
        field_key = field_keys[1] if len(field_keys) > 1 else None
        if not field_key:
            return must, must_not, should

        # ----- Check String ----
        if operator in ["StringEquals", "NumericEquals"]:
            if qualifier == "ForAnyValue":
                sh = [{"match": {field_key: value}} for value in values]
                should.extend(sh)
            else:
                mu = [{"match": {field_key: value}} for value in values]
                must.extend(mu)

        elif operator in ["StringNotEquals", "NumericNotEquals"]:
            if qualifier == "ForAnyValue":
                mu = [{"match": {field_key: value}} for value in values]
                must_not.extend(mu)
            else:
                sh = [{"match": {field_key: value}} for value in values]
                must_not.append({
                    "bool": {
                        "should": sh
                    }
                })
        elif operator == "StringContains":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: "*" + value + "*"}} for value in values]
                should.extend(sh)
            else:
                mu = [{"wildcard": {field_key: "*" + value + "*"}} for value in values]
                must.extend(mu)
        elif operator == "StringNotContains":
            if qualifier == "ForAnyValue":
                mu = [{"wildcard": {field_key: "*" + value + "*"}} for value in values]
                must_not.extend(mu)
            else:
                sh = [{"wildcard": {field_key: "*" + value + "*"}} for value in values]
                must_not.append({
                    "bool": {
                        "should": sh
                    }
                })
        elif operator == "StringEndsWith":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: "*" + value}} for value in values]
                should.extend(sh)
            else:
                mu = [{"wildcard": {field_key: "*" + value}} for value in values]
                must.extend(mu)

        elif operator == "StringStartsWith":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: value + "*"}} for value in values]
                should.extend(sh)
            else:
                mu = [{"wildcard": {field_key: value + "*"}} for value in values]
                must.extend(mu)
        elif operator == "StringMatchesRegex":
            if qualifier == "ForAnyValue":
                sh = [{"regexp": {field_key: value}} for value in values]
                should.extend(sh)
            else:
                mu = [{"regexp": {field_key: value}} for value in values]
                must.extend(mu)
        elif operator == "NumericLessThan":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lt": value}}} for value in values]
                must.extend(mu)
        elif operator == "NumericLessThanEquals":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lte": value}}} for value in values]
                must.extend(mu)
        elif operator == "NumericGreaterThan":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gt": value}}} for value in values]
                must.extend(mu)
        elif operator == "NumericGreaterThanEquals":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gte": value}}} for value in values]
                must.extend(mu)
        # Tùy data từng bên lưu datetime kiểu gì, Ví dụ Sale lưu theo timestamp thì cần convert ra start_time,
        # end_time trong ngày
        elif operator == "DateEquals":
            for value in values:
                start, end = ElasticsearchAdapter.convert_date_to_range_timestamp(value)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    should.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must.append({"range": {field_key: {"gte": start, "lte": end}}})
        elif operator == "DateNotEquals":
            sh = []
            for value in values:
                start, end = ElasticsearchAdapter.convert_date_to_range_timestamp(value)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    sh.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must_not.append({"range": {field_key: {"gte": start, "lte": end}}})
            if sh:
                must_not.append({
                    "bool": {"should": sh}
                })
        elif operator == "DateLessThan":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateLessThanEquals":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lte": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThan":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThanEquals":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gte": value}}} for value in values]
                must.extend(mu)

        # Check Day
        elif operator == "DayEquals":
            for value in values:
                start, end = ElasticsearchAdapter.convert_day_to_range_timestamp(value)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    should.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must.append({"range": {field_key: {"gte": start, "lte": end}}})
        elif operator == "DayNotEquals":
            sh = []
            for value in values:
                start, end = ElasticsearchAdapter.convert_day_to_range_timestamp(value)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    sh.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must_not.append({"range": {field_key: {"gte": start, "lte": end}}})
            if sh:
                must_not.append({
                    "bool": {"should": sh}
                })
        elif operator == "DayLessThan":
            values = ElasticsearchAdapter.convert_list_start_day_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateLessThanEquals":
            values = ElasticsearchAdapter.convert_list_start_day_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lte": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThan":
            values = ElasticsearchAdapter.convert_list_end_day_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThanEquals":
            values = ElasticsearchAdapter.convert_list_end_day_to_timestamp(values)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gte": value}}} for value in values]
                must.extend(mu)

        elif operator == "ListAllIn":
            for value in values:
                must.append({"terms": {field_key: value}})
        elif operator == "ListAllNotIn":
            must_not.append({"terms": {field_key: values}})
        elif operator == "ListAnyIn":
            for value in values:
                should.append({"terms": {field_key: value}})
        elif operator == "ListAnyNotIn":
            sh = [{"terms": {field_key: value}} for value in values]
            must_not.append({"bool": {"should": sh}})
        elif operator == "ListIsIn":
            must.append({"terms": {field_key: values}})
        elif operator == "ListIsNotIn":
            must_not.append({"terms": {field_key: values}})

        elif operator == "ListIsEmpty":
            must.append({
                "bool": {
                    "must_not": [
                        {
                            "exists": {
                                "field": field_key
                            }
                        }
                    ]
                }
            })
        elif operator == "ListIsNotEmpty":
            must.append({"exists": {"field": field_key}})

        return must, must_not, should

    @staticmethod
    def convert_date_to_range_timestamp(date):
        """
        :param date:
        :return:
        """
        try:
            date_input = datetime.strptime(date, "%Y-%m-%d %H:%M:%S")
            end_time = date_input.replace(second=59).timestamp()
            start_time = end_time - (60)
            return start_time, end_time
        except Exception:
            return None, None

    @staticmethod
    def convert_list_date_to_timestamp(lst_date):
        """
        :param lst_date:
        :return:
        """
        values = []
        for d in lst_date:
            try:
                value = datetime.strptime(d, "%Y-%m-%d %H:%M:%S").timestamp()
                values.append(value)
            except Exception:
                pass
        return values

    @staticmethod
    def convert_day_to_range_timestamp(date):
        """
        :param date:
        :return:
        """
        try:
            date_input = datetime.strptime(date, "%Y-%m-%d")
            end_time = date_input.replace(hour=17, minute=0, second=0).timestamp()
            start_time = end_time - (24 * 3600)
            return start_time, end_time
        except Exception:
            return None, None

    @staticmethod
    def convert_list_end_day_to_timestamp(lst_date):
        """
        :param lst_date:
        :return:
        """
        values = []
        for d in lst_date:
            try:
                value = datetime.strptime(d, "%Y-%m-%d").replace(hour=17, minute=59, second=59).timestamp()
                values.append(value)
            except Exception:
                pass
        return values

    @staticmethod
    def convert_list_start_day_to_timestamp(lst_date):
        """
        :param lst_date:
        :return:
        """
        values = []
        for d in lst_date:
            try:
                value = datetime.strptime(d, "%Y-%m-%d").replace(hour=17, minute=59, second=59).timestamp() - (
                            24 * 3600)
                values.append(value)
            except Exception:
                pass
        return values



"""
    Collection conditions
"""

from .all_in import AllInSchema
from .all_not_in import AllNotInSchema
from .any_in import AnyInSchema
from .any_not_in import AnyNotInSchema
from .is_empty import IsEmptySchema
from .is_in import IsInSchema
from .is_not_empty import IsNotEmptySchema
from .is_not_in import IsNotInSchema



"""
    Boolean conditions base class
"""

import logging

from marshmallow import Schema, fields, EXCLUDE

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)



class BooleanCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for Boolean conditions

        :param values: Boolean of values to compare during policy evaluation
    """

    def __init__(self, values, what, **kwargs):
        self.values = values
        self.what = what


    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is Boolean conditions satisfied

            :param what: Boolean to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()


class BooleanConditionSchema(Schema):
    """
        Base JSON schema class for Boolean conditions
    """
    values = fields.List(fields.Boolean(
        required=True,
        allow_none=False
    ), required=True,
        allow_none=False)
    what = fields.Boolean(
        required=True,
        allow_none=False
    )
    class Meta:
        unknown = EXCLUDE



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
               ..
              ( '`<
               )(
        ( ----'  '.
        (         ;
         (_______,' 
    ~^~^~^~^~^~^~^~^~^~^~
    Author: thong
    Company: M O B I O
    Date Created: 14/03/2023
"""


class DataAdapter:

    def build_query(self, _filter):
        raise NotImplementedError()

    def execute_query(self, _query):
        raise NotImplementedError()



"""
    String conditions base class
"""

import logging
import datetime
from marshmallow import Schema, fields, EXCLUDE, validate

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)


class DayCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for string conditions
    """

    class DateFormat:
        ISO_FORMAT = "%Y-%m-%d"

    def __init__(self, values, what, date_format=DateFormat.ISO_FORMAT, qualifier=ConditionBase.Qualifier.ForAnyValue, **kwargs):
        self.values = values
        self.what = what
        self.qualifier = qualifier
        self.date_format = date_format


    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is string conditions satisfied

            :param what: string value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()

    def convert_str_to_date(self, d_str):
        try:
            return datetime.datetime.strptime(d_str, self.DateFormat.ISO_FORMAT)
        except:
            return None

    def check_format_input_day(self, i):
        if isinstance(i, str):
            if self.convert_str_to_date(i):
                return True
        return False


class DayConditionSchema(Schema):
    """
        Base JSON schema for string conditions
    """
    values = fields.List(fields.Raw(required=True, allow_none=False), required=True, allow_none=False)
    what = fields.String(required=True, allow_none=False)
    qualifier = fields.String(allow_none=False, load_default=ConditionBase.Qualifier.ForAnyValue,
                            validate=validate.OneOf([ConditionBase.Qualifier.ALL]))
    date_format = fields.String(allow_none=False, load_default=DayCondition.DateFormat.ISO_FORMAT,)

    class Meta:
        unknown = EXCLUDE


"""
    String conditions base class
"""

import logging
import ipaddress
from marshmallow import Schema, fields, EXCLUDE, ValidationError

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)

def validate_ip_address(address):
    try:
        if isinstance(address, str):
            ipaddress.ip_address(address)
        else:
            for i in address:
                ipaddress.ip_address(i)
    except Exception:
        raise ValidationError("Invalid ip_address '{}'.".format(address))

def validate_ip_network(address):
    try:
        if isinstance(address, str):
            ipaddress.ip_network(address)
        else:
            for i in address:
                ipaddress.ip_network(i)
    except Exception:
        raise ValidationError("Invalid ip_network '{}'.".format(address))

class IpAddressCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for string conditions
    """

    def __init__(self, values, what, **kwargs):
        self.values = values
        self.what = what


    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is string conditions satisfied

            :param what: string value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()


class IpAddressConditionSchema(Schema):
    """
        Base JSON schema for string conditions
    """
    values = fields.List(fields.String(required=True, allow_none=False), required=True, allow_none=False, validate=validate_ip_network)
    what = fields.String(required=True, allow_none=False, validate=validate_ip_address)

    class Meta:
        unknown = EXCLUDE


"""
    Value is not in collection conditions
"""

from marshmallow import post_load

from .base import CollectionCondition, CollectionConditionSchema


class IsNotIn(CollectionCondition):
    """
        Condition for `what` is not a member of `values`
    """

    # def is_satisfied(self, ctx) -> bool:
    #     return self._is_satisfied(ctx.attribute_value)

    def _is_satisfied(self) -> bool:
        return self.what not in self.values


class IsNotInSchema(CollectionConditionSchema):
    """
        JSON schema for is not in collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return IsNotIn(**data)



from mobio.libs.caching import LruCache
import os, requests


class StoreCacheType:
    LOCAL = 1
    REDIS = 2

class Cache:
    REDIS_URI = "{}?health_check_interval=30".format(os.environ.get("ADMIN_REDIS_URI", os.environ.get("REDIS_URI")))
    PREFIX = "m_abac"

lru_cache_redis = LruCache(
    store_type=StoreCacheType.REDIS,
    redis_uri=Cache.REDIS_URI,
    cache_prefix=Cache.PREFIX,
)

class Mobio:
    ADMIN_HOST = os.environ.get("ADMIN_HOST")
    MOBIO_TOKEN = "Basic {}".format(os.environ.get('YEK_REWOP', ''))



class APIRequest:
    TimeOut = 10
    ADMIN_GET_FULL_INFO_ACCOUNT = "{domain}/adm/api/v2.1/accounts/{account_id}/full-info"
    ADMIN_GET_LIST_ACTION_FOR_MERCHANT = "{domain}/adm/api/v2.1/policies/actions"
    ADMIN_GET_LIST_STATEMENT = "{domain}/adm/api/v2.1/statements"

class CallAPI:

    @lru_cache_redis.add()
    @staticmethod
    def admin_get_account_info(merchant_id, account_id):
        try:
            url = APIRequest.ADMIN_GET_FULL_INFO_ACCOUNT.format(domain=Mobio.ADMIN_HOST, account_id=account_id)
            authorization = Mobio.MOBIO_TOKEN
            data_res = requests.get(
                url,
                headers={
                    "Authorization": authorization,
                    "X-Merchant-ID": merchant_id,
                },
                timeout=APIRequest.TimeOut,
            )
            data = None
            if data_res.status_code == 200:
                if data_res.json().get("data"):
                    data = data_res.json().get("data")
            # print("admin_get_account_info: {}".format(data))
            return data
        except Exception as er:
            err_msg = "admin_get_account_info err: {}".format(er)
            print(err_msg)
            return {}

    @lru_cache_redis.add()
    @staticmethod
    def admin_get_json_action(merchant_id):
        try:
            url = APIRequest.ADMIN_GET_LIST_ACTION_FOR_MERCHANT.format(domain=Mobio.ADMIN_HOST)
            authorization = Mobio.MOBIO_TOKEN
            data_res = requests.get(
                url,
                headers={
                    "Authorization": authorization,
                    "X-Merchant-ID": merchant_id,
                },
                timeout=APIRequest.TimeOut,
                params={"page": "-1"}
            )
            data = {}
            if data_res.status_code == 200:
                if data_res.json().get("data"):
                    list_data = data_res.json().get("data")
                    for i in list_data:
                        data[i.get("key")] = i
            # print("admin_get_json_action: {}".format(data))
            return data
        except Exception as er:
            err_msg = "admin_get_json_action err: {}".format(er)
            print(err_msg)
            return {}

    @staticmethod
    @lru_cache_redis.add(expiration=900)
    def admin_get_list_statement(merchant_id, account_id, resource, action, service):
        """
        :param merchant_id:
        :param account_id:
        :param resource:
        :param action:
        :return: [{
            "effect": "allow",
            "action": [
                "DealViewReport"
            ],
            "resource": [
                "sale:deal"
            ],
            "condition": [
                {
                    "operator": "StringEquals",
                    "field": "user:block",
                    "values": [
                        "KHDN"
                    ],
                    "qualifier": "ForAnyValue",
                    "if_exists": False,
                    "ignore_case": False
                },
                {
                    "operator": "StringEquals",
                    "field": "deal:block",
                    "values": [
                        "${user:block}"
                    ],
                    "qualifier": "ForAnyValue",
                    "if_exists": False,
                    "ignore_case": False
                },
                {
                    "operator": "StringStartsWith",
                    "field": "deal:scope_code",
                    "values": [
                        "${user:scope_code}"
                    ],
                    "qualifier": "ForAnyValue",
                    "if_exists": False,
                    "ignore_case": False
                }
            ]
        }]
        """
        try:
            url = APIRequest.ADMIN_GET_LIST_STATEMENT.format(domain=Mobio.ADMIN_HOST)
            authorization = Mobio.MOBIO_TOKEN
            body_req = {
                'user_ids': [account_id],
                'team_ids': [],
                'merchant_ids': [merchant_id],
                'resources': [service + ":" + resource],
                'actions': [resource + ":" + action]
            }
            data_res = requests.post(
                url,
                json=body_req,
                headers={
                    "Authorization": authorization,
                    "X-Merchant-ID": merchant_id,
                },
                timeout=APIRequest.TimeOut,
            )
            data = None
            if data_res.status_code == 200:
                if data_res.json().get("statements"):
                    data = data_res.json().get("statements")
            # print("admin_get_list_statement: {}".format(data))
            return data

            # return [
            #     {
            #         "effect": "allow",
            #         "action": [
            #             "ListFromSale"
            #         ],
            #         "resource": [
            #             "sale:deal"
            #         ],
            #         "condition": [
            #             {
            #                 "operator": "StringEquals",
            #                 "field": "deal:block",
            #                 "values": [
            #                     "${user:block}"
            #                 ],
            #                 "qualifier": "ForAnyValue",
            #                 "if_exists": False,
            #                 "ignore_case": False
            #             },
            #             {
            #                 "operator": "StringStartsWith",
            #                 "field": "deal:scope_code",
            #                 "values": [
            #                     "${user:scope_code}"
            #                 ],
            #                 "qualifier": "ForAnyValue",
            #                 "if_exists": False,
            #                 "ignore_case": False
            #             }
            #         ]
            #     },
            #             {
            #                 "effect": "allow",
            #                 "action": [
            #                     "ListFromSale"
            #                 ],
            #                 "resource": [
            #                     "sale:deal"
            #                 ],
            #                 "condition": [
            #                     {
            #                         "operator": "StringStartsWith",
            #                         "field": "deal:scope_code",
            #                         "values": [
            #                             "3#7", "3#8", "3#9"
            #                         ],
            #                         "qualifier": "ForAnyValue",
            #                         "if_exists": False,
            #                         "ignore_case": False
            #                     },
            #                     {
            #                         "operator": "StringEquals",
            #                         "field": "user:scope_code",
            #                         "values": [
            #                             "3#6"
            #                         ],
            #                         "qualifier": "ForAnyValue",
            #                         "if_exists": False,
            #                         "ignore_case": False
            #                     }
            #                 ]
            #             }
            # ]

        except Exception as er:
            err_msg = "admin_get_list_statement err: {}".format(er)
            print(err_msg)
            return []



"""
    All of the values in collection conditions
"""

from marshmallow import post_load

from .base import CollectionCondition, CollectionConditionSchema


class AllIn(CollectionCondition):
    """
        Condition for all values of `what` in `values`
    """

    def _is_satisfied(self) -> bool:
        return set(self.what).issubset(self.values)


class AllInSchema(CollectionConditionSchema):
    """
        JSON schema for all in collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return AllIn(**data)



"""
    String not contains conditions
"""

from marshmallow import post_load

from .base import StringCondition, StringConditionSchema


class NotContains(StringCondition):
    """
        Condition for string `self.what` not contains `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if self.case_insensitive:
                    if self.what.lower() in i.lower():
                        return False
                else:
                    if self.what in i:
                        return False
            return True
        else:
            for i in self.values:
                if self.case_insensitive:
                    if self.what.lower() in i.lower():
                        return False
                else:
                    if self.what in i:
                        return False
            return True


class NotContainsSchema(StringConditionSchema):
    """
        JSON schema for not contains string condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return NotContains(**data)



from copy import deepcopy
import re, json
from datetime import datetime
from user_agents import parse
from flask import request
import base64


class Utils:

    @staticmethod
    def split_delemiter_resource(str_value):
        resource_name, resource_key = None, None
        if ":" in str_value:
            values = str_value.split(":")
            resource_name, resource_key = values[0], values[1]
        else:
            resource_name = str_value
        return resource_name, resource_key

    @staticmethod
    def split_delemiter_field(str_value):
        return str_value.split(":")

    @staticmethod
    def get_nested_value(obj, field):
        obj_value = deepcopy(obj)
        fields = field.split('.')
        for f in fields:
            obj_value = obj_value.get(f)
            if obj_value is None:
                return None
        return obj_value

    @staticmethod
    def get_field_key_from_variable(str_text) -> list:
        key_regex = "\${(.*?)}"
        try:
            if isinstance(str_text, int):
                return []
            return re.findall(key_regex, str_text)
        except Exception as er:
            print("get_field_key_from_variable key: {}, err: {}".format(key_regex, er))
            return []

    @staticmethod
    def replace_variable_to_value(field_key, field_value, str_origin):
        key_regex = "\${" + field_key + "}"
        return re.sub(fr'{key_regex}', field_value, str_origin)

    @staticmethod
    def check_field_is_variable(field_key):
        if field_key.startswith("${") and field_key.endswith("}"):
            return True
        return False

    @staticmethod
    def get_date_utcnow():
        return datetime.utcnow()

    @staticmethod
    def parse_user_agent(ua_string):
        try:
            user_agent = parse(ua_string)
            # if user_agent.is_pc:
            #     device = "web"
            # else:
            #     device = "mobile"
            agent_info = {
                "env:browser": user_agent.browser,
                "env:os": user_agent.os,
                # "env:device_type": device,
            }
            return agent_info
        except:
            return {}

    @staticmethod
    def get_info_from_header_request():
        try:
            header_data = request.headers
            user_agent = header_data.get('User-Agent')
            current_ip = request.remote_addr
            if header_data.get("X-Real-Ip"):
                current_ip = header_data.get("X-Real-Ip")
            device_type = header_data.get('x-device-type')
            return {
                "env:source_ip": current_ip,
                "env:user_agent": user_agent,
                "env:device_type": device_type if device_type else "web"
            }
        except:
            return {}

    @staticmethod
    def get_fields_of_resource(full_resource):
        fields = {}
        if "/" not in full_resource:
            return fields
        split_resource = full_resource.split("/")
        service_resource, field_multi = split_resource[0], split_resource[1]
        resource = service_resource.split(":")[-1]
        return {resource: field_multi.split(",")}

    @staticmethod
    def field_in_body_request(list_field, body_request):
        if list_field and body_request:
            for field in list_field:
                if field in body_request:
                    return True
        return False

    @staticmethod
    def base64_encode(data):
        try:
            byte_string = data.encode('utf-8')
            encoded_data = base64.b64encode(byte_string)
            return encoded_data.decode(encoding='UTF-8')
        except:
            return ""

    @staticmethod
    def base64_decode(encoded_data):
        try:
            if isinstance(encoded_data, bytes):
                encoded_data = encoded_data.decode('UTF-8')
            decoded_data = base64.urlsafe_b64decode(encoded_data + '=' * (-len(encoded_data) % 4))
            return decoded_data.decode(encoding='UTF-8')
        except:
            return encoded_data

    @staticmethod
    def get_info_jwt():
        try:
            data_author = request.headers['Authorization']
            if not data_author:
                raise ValueError("Authorization not found")
            auth_type, jwt_token = data_author.split(None, 1)
            auth_type = auth_type.lower()
            if auth_type not in ['basic', 'bearer', 'digest']:
                raise ValueError("Authorization invalid")
            json_token = {}
            if auth_type == 'bearer':
                arr_token = jwt_token.split(".")
                data = Utils.base64_decode(arr_token[1].replace('_', '/') + '===')
                body_token = json.loads(data.decode('utf8'))
                merchant_id = body_token.get('merchant_id', None)
                account_id = body_token.get('id', None)
                json_token.update({
                    "merchant_id": merchant_id,
                    "account_id": account_id,
                })
            return auth_type, json_token
        except:
            raise ValueError("Authorization invalid")



"""
    Condition one-of schema
"""
from marshmallow import EXCLUDE
from marshmallow_oneofschema import OneOfSchema
from .boolean import *
from .collection import *
from .date_time import *
from .ip_address import *
from .numeric import *
from .string import *
from .day import *
from .others import *

class ConditionSchema(OneOfSchema):
    """
        Polymorphic JSON schema for conditions
    """
    type_field = "operator"
    type_schemas = {
        "StringEquals": EqualsSchema,
        "StringNotEquals": NotEqualsSchema,
        "StringContains": ContainsSchema,
        "StringNotContains": NotContainsSchema,
        "StringEndsWith": EndsWithSchema,
        "StringStartsWith": StartsWithSchema,
        "StringMatchesRegex": RegexMatchSchema,
        "StringNotEndsWith": NotEndsWithSchema,
        "StringNotStartsWith": NotStartsWithSchema,

        "NumericEquals": EqSchema,
        "NumericNotEquals": NeqSchema,
        "NumericLessThan": LtSchema,
        "NumericLessThanEquals": LteSchema,
        "NumericGreaterThan": GtSchema,
        "NumericGreaterThanEquals": GteSchema,

        "DateEquals": DateEqSchema,
        "DateNotEquals": DateNeqSchema,
        "DateLessThan": DateLtSchema,
        "DateLessThanEquals": DateLteSchema,
        "DateGreaterThan": DateGtSchema,
        "DateGreaterThanEquals": DateGteSchema,

        "Bool": CheckBoolSchema,

        "IpAddress": InSubnetSchema,
        "NotIpAddress": NotInSubnetSchema,

        "ListAllIn": AllInSchema,
        "ListAllNotIn": AllNotInSchema,
        "ListAnyIn": AnyInSchema,
        "ListAnyNotIn": AnyNotInSchema,
        "ListIsEmpty": IsEmptySchema,
        "ListIsIn": IsInSchema,
        "ListIsNotEmpty": IsNotEmptySchema,
        "ListIsNotIn": IsNotInSchema,

        "DayEquals": DayEqSchema,
        "DayNotEquals": DayNeqSchema,
        "DayLessThan": DayLtSchema,
        "DayLessThanEquals": DayLteSchema,
        "DayGreaterThan": DayGtSchema,
        "DayGreaterThanEquals": DayGteSchema,

        "Exists": ExistsSchema,
        "NotExists": NotExistsSchema,

    }

    class Meta:
        unknown = EXCLUDE




"""
    String contains conditions
"""

from marshmallow import post_load
import ipaddress
from .base import IpAddressCondition, IpAddressConditionSchema


class InSubnet(IpAddressCondition):
    """
        Condition for string `self.what` contains `value`
    """

    def _is_satisfied(self) -> bool:
        for i in self.values:
            if ipaddress.ip_address(self.what) in ipaddress.ip_network(i):
                return True
        return False

class InSubnetSchema(IpAddressConditionSchema):
    """
        JSON schema for contains string conditions
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return InSubnet(**data)



"""
    Attribute exists conditions
"""
from marshmallow import post_load

from .base import OthersCondition, OthersConditionSchema


class Exists(OthersCondition):
    """
        Condition for attribute value exists
    """

    def is_satisfied(self) -> bool:
        if self.values.get(self.what) or self.values.get(self.what) in [0, False]:
            return True
        else:
            return False


class ExistsSchema(OthersConditionSchema):
    """
        JSON schema for attribute value exists conditions
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return Exists(**data)




class ErrorCodeException:
    GET_VALUE_NONE = 1001
    CONDITION_INVALID = 1002
    CONDITION_EMPTY = 1003
    CHECK_POLICY = 1004
    REQUEST_ACCESS_CREATE = 1005

class BaseException(Exception):
    def __init__(self, error_code, message):
        self.error_code = error_code
        self.message = message
        super().__init__(self.message)

class GetValueNoneException(BaseException):
    def __init__(self, message):
        error_code = ErrorCodeException.GET_VALUE_NONE
        message = " - " + str(error_code) + ": {}".format(message)
        super().__init__(error_code, message)

class InvalidConditionException(BaseException):
    def __init__(self, message):
        error_code = ErrorCodeException.CONDITION_INVALID
        message = " - " + str(error_code) + ": {}".format(message)
        super().__init__(error_code, message)

class EmptyConditionException(BaseException):
    def __init__(self, message):
        error_code = ErrorCodeException.CONDITION_EMPTY
        message = " - " + str(error_code) + ": {}".format(message)
        super().__init__(error_code, message)

class CheckPolicyException(BaseException):
    def __init__(self, message):
        error_code = ErrorCodeException.CHECK_POLICY
        message = " - " + str(error_code) + ": {}".format(message)
        super().__init__(error_code, message)

class RequestAccessCreateException(BaseException):
    def __init__(self, message):
        error_code = ErrorCodeException.REQUEST_ACCESS_CREATE
        message = " - " + str(error_code) + ": {}".format(message)
        super().__init__(error_code, message)


class ResultAccess:
    def __init__(self):
        self.allow_access = False
        self.display_config = []
        self.filter_config = []

    def get_allow_access(self):
        return self.allow_access

    def set_allow_access(self, value: bool):
        self.allow_access = value

    def get_filter_config(self):
        """

        filter: [{'effect': 'allow', 'condition': [{'operator': 'StringEquals', 'field': 'deal:block',
        'values': ['KHDN'], 'qualifier': 'ForAnyValue', 'if_exists': False, 'ignore_case': False},
        {'operator': 'StringStartsWith', 'field': 'deal:scope_code', 'values': ['MB##HN'],
        'qualifier': 'ForAnyValue', 'if_exists': False, 'ignore_case': False}]}]
        :return:
        """
        # print(self.filter_config)
        dict_field = {}
        dict_exists = {}
        list_filter_config = []
        index_statement = 0
        for statement in self.filter_config:
            item_condition = []
            index_condition = 0
            effect = statement.get("effect")
            for condition in statement.get("condition"):
                operator = condition.get("operator")
                field = condition.get("field")
                qualifier = condition.get("qualifier")
                key_filter = self.build_key_filter_config(effect, operator, field, qualifier)
                if key_filter not in dict_field:
                    dict_field[key_filter] = [index_statement, index_condition]
                    item_condition.append(condition)
                else:
                    dict_exists[key_filter] = condition
                index_condition += 1
            if item_condition:
                list_filter_config.append({
                    'effect': effect,
                    'condition': item_condition
                })
            index_statement += 1
        for k, v in dict_exists.items():
            if k in dict_field:
                list_filter_config[dict_field.get(k)[0]].get("condition")[dict_field.get(k)[1]].get("values").extend(v.get("values"))

        return list_filter_config

    def add_filter_config(self, value):
        self.filter_config.append(value)

    def get_display_config(self):
        return self.display_config

    def add_display_config(self, value):
        self.display_config.append(value)

    @classmethod
    def build_key_filter_config(cls, effect, operator, field, qualifier):
        return effect + "#" + operator + "#" + field + "#" + qualifier

    @classmethod
    def extract_key_filter_config(cls, key_filter):
        items = key_filter.split("#")
        return {
            "effect": items[0],
            "operator": items[1],
            "field": items[2],
            "qualifier": items[3],
        }



"""
    Any of the values not in collection conditions
"""

from marshmallow import post_load

from .base import CollectionCondition, CollectionConditionSchema


class AnyNotIn(CollectionCondition):
    """
        Condition for any values of `what` not in `values`
    """

    def _is_satisfied(self) -> bool:
        return not bool(set(self.what).intersection(self.values))


class AnyNotInSchema(CollectionConditionSchema):
    """
        JSON schema for any not in collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return AnyNotIn(**data)



"""
    Others conditions base class
"""

import logging

from marshmallow import Schema, fields, EXCLUDE, validate

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)



class OthersCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for numeric conditions
    """

    def __init__(self, values, what, **kwargs):
        self.values = values
        self.what = what


    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is numeric conditions satisfied

            :param what: numeric value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()


class OthersConditionSchema(Schema):
    """
        Base JSON schema for numeric conditions
    """
    values = fields.Dict(default={}, missing={}, allow_none=False)
    what = fields.String(required=True, allow_none=False)

    class Meta:
        unknown = EXCLUDE


"""
    Policy decision point implementation
"""

from copy import deepcopy

from .call_api import CallAPI
from .policy import PolicySchema, AccessType
from .policy.utils import Utils
from .result_access import ResultAccess


class PolicyDecisionPoint(object):
    """
        Policy decision point
    """

    class AccountType:
        SYSTEM = "system"
        NORMAL = "normal"

    class AccessLevel:
        LIST = "list"
        READ = "read"
        ADD = "add"
        EDIT = "edit"
        DELETE = "delete"
        OTHER = "other"

    class AuthorizationType:
        BASIC = 'basic'
        BEARER = 'bearer'
        DIGEST = 'digest'

    def __init__(self,
                 merchant_id: str = None, resource: str = None, action: str = None, account_id: str = None,
                 user_info: dict = None, data_before: dict = None, data_after: dict = None, environment: dict = None,
                 account_type=AccountType.NORMAL):
        if not resource:
            raise ValueError("resource required")
        if not action:
            raise ValueError("action required")
        if not merchant_id or not account_id:
            auth_type, json_token = Utils.get_info_jwt()
            if auth_type == self.AuthorizationType.BEARER:
                merchant_id = json_token.get("merchant_id")
                account_id = json_token.get("account_id")
            else:
                account_type = self.AccountType.SYSTEM
        if account_type == PolicyDecisionPoint.AccountType.NORMAL:
            if not merchant_id:
                raise ValueError("merchant_id required")
            if not account_id:
                raise ValueError("account normal account_id required")
            if not user_info:
                user_info = CallAPI.admin_get_account_info(merchant_id, account_id)

        self.account_type = account_type
        self.merchant_id = merchant_id
        self.account_id = account_id
        self.resource = resource
        self.action = action
        self.request_access = {}
        self.result_access = ResultAccess()
        self.request_access.update({
            "user": user_info if user_info else {},
            "env": self.get_info_environment(environment),
            self.resource: data_before if data_before else {}
        })
        self.data_after = data_after if data_after else {}
        self.data_before = data_before if data_before else {}
        self.access_level = ""
        self.service = ""

    def get_policy_statement_for_target(self):
        return CallAPI.admin_get_list_statement(self.merchant_id, self.account_id,
                                                self.resource, self.action, self.service)

    def is_allowed(self):
        """
            Check if authorization request is allowed

            :param request: request object
            :return: True if authorized else False
        """
        try:
            if self.account_type == PolicyDecisionPoint.AccountType.SYSTEM:
                self.result_access.set_allow_access(True)
                return self.result_access
            action_info = CallAPI.admin_get_json_action(merchant_id=self.merchant_id)
            if not action_info:
                raise ValueError("action for merchant_id {} not found".format(self.merchant_id))
            if not action_info.get(self.action):
                raise ValueError("action {} not found".format(self.action))
            self.access_level = action_info.get(self.action, {}).get("access_level", "").lower()
            if not self.access_level:
                raise ValueError("access_level {} not valid".format(self.action))
            self.service = action_info.get(self.action, {}).get("service", "")
            if not self.service:
                raise ValueError("service {} not valid".format(self.service))

            list_statement = self.get_policy_statement_for_target()
            if list_statement:

                if self.access_level == PolicyDecisionPoint.AccessLevel.ADD:
                    self.data_after.update(self.data_before)
                statement_check_allow, statement_check_deny = self.get_statement_by_type(list_statement)
                if statement_check_deny:
                    if self.check_list_statement_is_deny(statement_check_deny):
                        self.result_access.set_allow_access(False)
                        return self.result_access
                if statement_check_allow:
                    self.result_access.set_allow_access(self.check_list_statement_is_allow(statement_check_allow))
                else:
                    self.result_access.set_allow_access(False)
            else:
                self.result_access.set_allow_access(False)
        except Exception as err:
            print("Exception: {}".format(err))
            self.result_access.set_allow_access(False)

        return self.result_access

    def get_info_environment(self, environment):
        if not environment:
            environment = Utils.get_info_from_header_request()
            if environment:
                if environment.get("env:user_agent"):
                    environment.update(Utils.parse_user_agent(environment.get("env:user_agent")))
        environment = environment if environment else {}
        environment.update({"env:current_time": Utils.get_date_utcnow()})
        return environment

    def get_value_from_variable(self, str_variable):
        variables = Utils.get_field_key_from_variable(str_variable)
        if variables:
            if len(variables) == 1 and Utils.check_field_is_variable(variables[0]):
                field_value = self.get_value_from_field(variables[0])
                return field_value
            for field_key in variables:
                field_value = self.get_value_from_field(field_key)
                if field_value is not None:
                    field_value = str(field_value)
                    str_variable = Utils.replace_variable_to_value(field_key, field_value, str_variable)
        return str_variable

    def get_value_from_field(self, field_key):
        resource_name, resource_key = Utils.split_delemiter_resource(field_key)
        if resource_name and resource_key:
            if self.request_access.get(resource_name) and isinstance(self.request_access.get(resource_name), dict):
                data_resource = self.request_access.get(resource_name)
                value = Utils.get_nested_value(data_resource, resource_key)
                return value
        return None

    def get_statement_by_type(self, list_statement: list):
        statement_check_allow, statement_check_deny = [], []
        for statement in list_statement:
            statement_copy = deepcopy(statement)
            fields = []
            if statement_copy.get("resource_field") and statement_copy.get("resource_field").get(self.resource):
                fields = statement_copy.get("resource_field").get(self.resource)
            if fields:
                statement_copy["fields"] = fields
            if self.access_level == PolicyDecisionPoint.AccessLevel.LIST:
                condition_check = []
                condition_filter = []
                for item_cond in statement_copy.get("condition"):
                    if item_cond.get("field").startswith("user") or item_cond.get("field").startswith("env"):
                        condition_check.append(item_cond)
                    else:
                        values = [self.get_value_from_variable(i) for i in item_cond.get("values")]
                        item_cond["values"] = values
                        condition_filter.append(item_cond)

                if statement_copy.get("effect") == AccessType.ALLOW_ACCESS:
                    # chỉ xử lý các điều kiện bộ lọc với effect allow
                    if condition_filter:
                        statement_copy["condition_filter"] = condition_filter
                    statement_check_allow.append({
                        **statement_copy,
                        "condition": condition_check
                    })
                else:
                    statement_check_deny.append({
                        **statement_copy,
                        "condition": condition_check
                    })

            elif self.access_level == PolicyDecisionPoint.AccessLevel.READ:
                if statement_copy.get("effect") == AccessType.ALLOW_ACCESS:
                    statement_check_allow.append(statement_copy)
                else:
                    statement_check_deny.append(statement_copy)

            else:
                if fields:
                    statement_copy["check_field"] = 1
                if statement_copy.get("effect") == AccessType.ALLOW_ACCESS:
                    statement_check_allow.append(statement_copy)
                else:
                    statement_check_deny.append(statement_copy)
        return statement_check_allow, statement_check_deny

    def check_list_statement_is_deny(self, list_statement: list):
        result_deny = False
        try:
            for statement in list_statement:
                statement["request_access"] = self.request_access
                policy_schema = PolicySchema().load(statement)
                if policy_schema.is_allowed():
                    if self.access_level == PolicyDecisionPoint.AccessLevel.READ:
                        # tìm tất cả các cấu hình ẩn hiện field nếu có
                        if statement.get("fields"):
                            self.result_access.add_display_config({
                                "effect": statement.get("effect"),
                                "fields": statement.get("fields"),
                            })
                            continue
                    else:
                        # nếu ko có field nào ở cấu hình policy nằm trong dữ liệu gửi lên thì cho qua
                        if statement.get("check_field") and statement.get("fields"):
                            if not Utils.field_in_body_request(statement.get("fields"), self.data_after):
                                continue
                    result_deny = True
                    break
        except Exception as err:
            print("check_list_statement_is_deny err: {}".format(err))
            result_deny = False
        return result_deny

    def check_list_statement_is_allow(self, list_statement: list):
        result_allow = False
        try:
            for statement in list_statement:
                statement["request_access"] = self.request_access
                policy_schema = PolicySchema().load(statement)
                if policy_schema.is_allowed():
                    result_allow = True
                    if self.access_level == PolicyDecisionPoint.AccessLevel.READ:
                        # tìm tất cả các cấu hình ẩn hiện field nếu có
                        if statement.get("fields"):
                            self.result_access.add_display_config({
                                "effect": statement.get("effect"),
                                "fields": statement.get("fields"),
                            })
                            continue
                    elif self.access_level == PolicyDecisionPoint.AccessLevel.LIST:
                        # tìm tất cả các điều kiện bộ lọc với action list
                        if statement.get("condition_filter"):
                            self.result_access.add_filter_config(
                                {"effect": statement.get("effect"),
                                 "condition": statement.get("condition_filter")})
                            continue
                    else:
                        # kiểm tra thông tin field dữ liệu gửi lên với cấu hình field trong policy,
                        # nếu field gửi lên ko nằm trong tập con của cấu hình field thì reject statement
                        if statement.get("check_field") and statement.get("fields"):
                            list_field_data_after = list(self.data_after.keys())
                            if not set(list_field_data_after).issubset(statement.get("fields")):
                                result_allow = False
                                continue
                        break

        except Exception as err:
            print("check_list_statement_is_allow err: {}".format(err))
            result_allow = False
        return result_allow



"""
    String contains conditions
"""

from marshmallow import post_load

from .base import StringCondition, StringConditionSchema


class Contains(StringCondition):
    """
        Condition for string `self.what` contains `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if self.case_insensitive:
                    if self.what.lower() in i.lower():
                        return True
                else:
                    if self.what in i:
                        return True
            return False
        else:
            for i in self.values:
                if self.case_insensitive:
                    if self.what.lower() not in i.lower():
                        return False
                else:
                    if self.what not in i:
                        return False
            return True

class ContainsSchema(StringConditionSchema):
    """
        JSON schema for contains string conditions
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return Contains(**data)



"""
    String conditions
"""

from .contains import ContainsSchema
from .ends_with import EndsWithSchema
from .equals import EqualsSchema
from .not_contains import NotContainsSchema
from .not_equals import NotEqualsSchema
from .regex_match import RegexMatchSchema
from .starts_with import StartsWithSchema
from .not_starts_with import NotStartsWithSchema
from .not_ends_with import NotEndsWithSchema



"""
    Attribute does not exists conditions
"""

from marshmallow import post_load

from .base import OthersCondition, OthersConditionSchema


class NotExists(OthersCondition):
    """
        Condition for attribute value not exists
    """

    def is_satisfied(self) -> bool:
        if self.values.get(self.what) or self.values.get(self.what) in [0, False]:
            return False
        else:
            return True


class NotExistsSchema(OthersConditionSchema):
    """
        JSON schema for attribute value not exists condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use,unused-argument
        return NotExists(**data)



"""
    All of the values in collection conditions
"""

from marshmallow import post_load

from .base import BooleanCondition, BooleanConditionSchema
truthy = {
    "t",
    "T",
    "true",
    "True",
    "TRUE",
    "on",
    "On",
    "ON",
    "y",
    "Y",
    "yes",
    "Yes",
    "YES",
    "1",
    1,
    True,
}
falsy = {
    "f",
    "F",
    "false",
    "False",
    "FALSE",
    "off",
    "Off",
    "OFF",
    "n",
    "N",
    "no",
    "No",
    "NO",
    "0",
    0,
    0.0,
    False,
}

class CheckBool(BooleanCondition):
    """
        Condition for all values of `what` in `values`
    """

    def _is_satisfied(self) -> bool:
        for i in self.values:
            if (i in truthy and self.what in truthy) or (i in falsy and self.what in falsy):
                return True
        return False

class CheckBoolSchema(BooleanConditionSchema):
    """
        JSON schema for all in collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return CheckBool(**data)



"""
    Policy class
"""
import json
from marshmallow import Schema, fields, post_load, ValidationError, validate, EXCLUDE
from .conditions.schema import ConditionSchema
from .utils import Utils
from .conditions.base import ConditionBase
from .exceptions import *

class AccessType:
    DENY_ACCESS = "deny"
    ALLOW_ACCESS = "allow"

class Policy(object):
    """
        Policy class containing rules and targets
    """

    def __init__(
            self,
            effect: str,
            action: list,
            resource: list,
            condition: dict,
            request_access: dict, **kwargs
    ):
        self.effect = effect
        self.action = action
        self.resource = resource
        self.condition = condition
        self.request_access = request_access

    def is_allowed(self):
        conditions = self.convert_condition()
        for condition in conditions:
            # print("condition: ", condition)
            if condition.get("operator") == "Null":
                what_check = self.get_value_from_field(condition.get("field"))
                if what_check is not None:
                    return False
            else:
                cond_schema = ConditionSchema().load(condition)
                if not cond_schema._is_satisfied():
                    return False
        return True

    def convert_condition(self):
        """
            {
                "operator": "StringEquals",
                "field": "user:staff_code",
                "values": ["A123456"],
                "qualifier": "ForAnyValue",
                "if_exists": false,
                "ignore_case": false,
            }
        :return:
        """
        list_condition = []
        have_condition_exclude = False
        for item in self.condition:
            cond_schema = self.validate_item_condition(item)
            # print("cond_schema: {}".format(json.dumps(cond_schema)))
            if cond_schema.get("operator") == "Null":
                list_condition.append({**item})
            else:
                if_exists = cond_schema.get("if_exists")
                what_check = self.get_value_from_field(cond_schema.get("field"))
                if what_check is None:
                    if if_exists:
                        have_condition_exclude = True
                        continue
                    raise GetValueNoneException("{} get value is None".format(cond_schema.get("field")))
                values = []
                for v in cond_schema.get("values"):
                    values.append(self.get_value_from_variable(v))
                list_condition.append({
                    **item,
                    "values": values,
                    "what": what_check,
                })
        # if len(list_condition) == 0 and not have_condition_exclude:
        #     raise EmptyConditionException("no condition found")
        return list_condition

    def get_value_from_variable(self, str_variable):
        variables = Utils.get_field_key_from_variable(str_variable)
        if variables:
            if len(variables) == 1 and Utils.check_field_is_variable(variables[0]):
                field_value = self.get_value_from_field(variables[0])
                if field_value is None:
                    raise GetValueNoneException("{} get value is None".format(variables[0]))
                return field_value
            for field_key in variables:
                field_value = self.get_value_from_field(field_key)
                if field_value is None:
                    raise GetValueNoneException("{} get value is None".format(field_key))
                field_value = str(field_value)
                str_variable = Utils.replace_variable_to_value(field_key, field_value, str_variable)
        return str_variable

    def get_value_from_field(self, field_key):
        resource_name, resource_key = Utils.split_delemiter_resource(field_key)
        if resource_name and resource_key:
            if self.request_access.get(resource_name) and isinstance(self.request_access.get(resource_name), dict):
                data_resource = self.request_access.get(resource_name)
                value = Utils.get_nested_value(data_resource, resource_key)
                return value
        return None

    def validate_item_condition(self, obj):
        try:
            return ConditionValidateSchema().load(obj)
        except Exception as err:
            raise InvalidConditionException("validate_item_condition: {}".format(err))

class ConditionValidateSchema(Schema):
    operator = fields.String(required=True, allow_none=False)
    field = fields.String(required=True, allow_none=False)
    values = fields.List(fields.Raw(required=True, allow_none=False), required=True, allow_none=False)
    qualifier = fields.String(default=ConditionBase.Qualifier.ForAnyValue,  validate=validate.OneOf(ConditionBase.Qualifier.ALL))
    if_exists = fields.Boolean(default=False)
    ignore_case = fields.Boolean(default=False)

    class Meta:
        unknown = EXCLUDE


class PolicySchema(Schema):
    """
        JSON schema for policy
    """
    effect = fields.String(required=True, validate=validate.OneOf([AccessType.DENY_ACCESS, AccessType.ALLOW_ACCESS]))
    action = fields.List(fields.String(required=True, allow_none=False),required=True, allow_none=False)
    resource = fields.List(fields.String(required=True, allow_none=False),required=True, allow_none=False)
    condition = fields.List(fields.Raw(required=True, allow_none=False), required=True, allow_none=False)
    request_access = fields.Dict(default={}, missing={}, allow_none=False)

    class Meta:
        unknown = EXCLUDE

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        return Policy(**data)



"""
    Datetime conditions
"""

from .date_eq import DateEqSchema
from .date_gt import DateGtSchema
from .date_gte import DateGteSchema
from .date_lt import DateLtSchema
from .date_lte import DateLteSchema
from .date_neq import DateNeqSchema




from marshmallow import post_load
from .base import DayCondition, DayConditionSchema


class DayGt(DayCondition):

    def _is_satisfied(self) -> bool:
        if not self.check_format_input_day(self.what):
            return False
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if self.check_format_input_day(i):
                    if self.what > i:
                        return True
            return False
        else:
            for i in self.values:
                if not self.check_format_input_day(i):
                    return False
                if self.what <= i:
                    return False
            return True


class DayGtSchema(DayConditionSchema):
    """
        JSON schema for greater than datetime condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return DayGt(**data)
